# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QComboBox, QFormLayout, QLineEdit

from ...helpers import BaseDialog
from ...shell import EventType

from ..helpers import tagged_items

from .helpers import init_version_selector, validate_version_name


class RenameVersionDialog(BaseDialog):
    """ This class implements the dialog for renaming a version. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        self._version = QComboBox()
        layout.addWidget(self._version)

        form = QFormLayout()
        layout.addLayout(form)

        self._new_name = QLineEdit()
        form.addRow("New name", self._new_name)

    def set_fields(self):
        """ Set the dialog's fields from the project. """

        init_version_selector(self._version, self.model)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        old_name = self._version.currentText()

        new_name = self._new_name.text().strip()
        if not validate_version_name(new_name, project, self):
            return False

        # Rename in each API item it appears.
        for api_item, _ in tagged_items(project):
            for v in api_item.versions:
                if v.startversion == old_name:
                    v.startversion = new_name

                if v.endversion == old_name:
                    v.endversion = new_name

        # Rename in the header file versions.
        for hdir in project.headers:
            for hfile in hdir.content:
                for hfile_version in hfile.versions:
                    if hfile_version.version == old_name:
                        hfile_version.version = new_name

        # Rename in the project's list.
        project.versions[project.versions.index(old_name)] = new_name

        self.shell.notify(EventType.VERSION_RENAME, (old_name, new_name))

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass, field

from .header_directory import HeaderDirectory
from .module import Module
from .project_version import ProjectVersion


@dataclass
class Project:
    """ This class implements a project. """

    # The name of the project file.  Note that this isn't part of the project
    # file itself.
    name: str = ''

    # The dirty state of the project.  Note that this isn't part of the project
    # file itself.
    dirty: bool = False

    # The list of externally defined features.
    externalfeatures: list[str] = field(default_factory=list)

    # The list of externally defined modules.
    externalmodules: list[str] = field(default_factory=list)

    # The list of features.
    features: list[str] = field(default_factory=list)

    # The list of header directories.
    headers: list[HeaderDirectory] = field(default_factory=list)

    # The list of Python modules.
    modules: list[Module] = field(default_factory=list)

    # The list of platforms.
    platforms: list[str] = field(default_factory=list)

    # The name of the optional root Python module.
    rootmodule: str = ''

    # The comments placed at the start of every generated .sip file.
    sipcomments: str = ''

    # The version number of the project format.
    version: tuple[int] = ProjectVersion

    # The ordered list of versions.
    versions: list[str] = field(default_factory=list)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ...helpers import version_range

from ..version_range import VersionRange

from .base_adapter import AttributeType, BaseAdapter


class TaggedAdapter(BaseAdapter):
    """ This is the Tagged adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'features':     AttributeType.STRING_LIST,
        'platforms':    AttributeType.STRING_LIST,
    }

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        # Load the simple attributes.
        super().load(element, project, ui)

        versions = element.get('versions')

        if versions is not None:
            for version in versions.split():
                version_range = VersionRange()
                version_range.startversion, version_range.endversion = version.split('-')
                self.model.versions.append(version_range)

    def save_attributes(self, output):
        """ Save the XML attributes. """

        versions = self.versions_as_str(as_xml=True)
        if versions != '':
            self.save_attribute('versions', versions, output)

        self.save_str_list('platforms', output)
        self.save_str_list('features', output)

    def versions_as_str(self, as_xml=False):
        """ Return the standard string representation of the versions. """

        version_ranges = []

        for version in self.model.versions:
            version_range_s = version_range(version)

            if as_xml:
                version_range_s = version_range_s.replace(' ', '')

            version_ranges.append(version_range_s)

        separator = ' ' if as_xml else ', '

        return separator.join(version_ranges)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..annos import Annos
from ..code import Code
from ..code_container import CodeContainer
from ..enum import Enum
from ..function import Function
from ..klass import Class
from ..manual_code import ManualCode
from ..namespace import Namespace
from ..opaque_class import OpaqueClass
from ..operator_function import OperatorFunction
from ..typedef import Typedef
from ..variable import Variable

from .adapt import adapt
from .base_adapter import AttributeType, BaseApiAdapter


class NamespaceAdapter(BaseApiAdapter):
    """ This is the Namespace adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'name': AttributeType.STRING,
    }

    # The map of element tags and Code sub-class factories.
    _TAG_CODE_MAP = {
        'Class':            Class,
        'Enum':             Enum,
        'Function':         Function,
        'ManualCode':       ManualCode,
        'Namespace':        Namespace,
        'OpaqueClass':      OpaqueClass,
        'OperatorFunction': OperatorFunction,
        'Typedef':          Typedef,
        'Variable':         Variable,
    }

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        namespace = self.model
        other_namespace = other.model

        if type(namespace) is not type(other_namespace):
            return False

        if namespace.name != other_namespace.name:
            return False

        return True

    def as_str(self):
        """ Return the standard string representation. """

        namespace = self.model

        return 'namespace ' + namespace.name + adapt(namespace, Annos).as_str()

    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        namespace = self.model

        nr_ends = self.version_start(output)

        output.blank()

        output.write(self.as_str())
        output.write('\n{\n')

        output.write('%TypeHeaderCode\n', indent=False)

        if namespace.typeheadercode != '':
            output.write(namespace.typeheadercode + '\n', indent=False)
        else:
            output.write(f'#include <{sip_file.name}>\n', indent=False)

        output.write('%End\n', indent=False)

        output.blank()

        output += 1

        for api in namespace.content:
            if api.status == '':
                adapt(api).generate_sip(sip_file, output)

        output -= 1

        output.write('};\n')

        output.blank()

        self.version_end(nr_ends, output)

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        adapt(self.model, Code).load(element, project, ui)
        adapt(self.model, CodeContainer).load(self._TAG_CODE_MAP, element,
                project, ui)

    def save(self, output):
        """ Save the model to an output file. """

        namespace = self.model

        output.write('<Namespace')
        adapt(namespace, Code).save_attributes(output)
        adapt(namespace, CodeContainer).save_attributes(output)
        self.save_attribute('name', namespace.name, output)
        output.write('>\n')

        output += 1
        self.save_literal('typeheadercode', output)
        adapt(namespace, Code).save_subelements(output)
        adapt(namespace, CodeContainer).save_subelements(output)
        output -= 1

        output.write('</Namespace>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from abc import ABC, abstractmethod
from enum import auto, Enum
from xml.sax.saxutils import escape

from ...helpers import version_range

from .adapt import adapt


class AttributeType(Enum):
    """ The different types of element and model attributes. """

    BOOL = auto()
    LITERAL = auto()
    STRING = auto()
    STRING_LIST = auto()


class BaseAdapter(ABC):
    """ This is the base class for all adapters and provides the ability to
    load and save a model to a project file and to provide a user-friendly, one
    line string representation.
    """

    # The default attribute type map.
    ATTRIBUTE_TYPE_MAP = {}

    def __init__(self, model):
        """ Initialise the adapter. """

        self.model = model

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        # This method must be reimplemented by those adapters that contribute
        # to the comparison of two APIs.  However we don't want to make it
        # abstract and have to provide a stub reimplementation in other
        # adapters.
        raise NotImplementedError

    def as_str(self):
        """ Return the standard string representation. """

        # This method must be reimplemented by those adapters that contribute
        # to the string representation of an API.  However we don't want to
        # make it abstract and have to provide a stub reimplementation in other
        # adapters.
        raise NotImplementedError

    @classmethod
    def expand_type(cls, type, name=None):
        """ Return the full type with an optional name. """

        # Handle the trivial case.
        if type == '':
            return ''

        # This is entirely cosmetic to be consistent with older versions.
        type = cls._normalise_templates(type)

        # SIP can't yet handle every C++ fundamental type.
        s = type.replace('long int', 'long')

        # If there is no embedded %s then just append the name.
        if '%s' in s:
            s = s % name
        elif name:
            if s[-1] not in '&*':
                s += ' '

            s += name

        return s

    def generate_sip_directives(self, output):
        """ Write any directives to a .sip file. """

        # This default implementation does nothing.
        pass

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        # This default implementation loads attributes define by
        # ATTRIBUTE_TYPE_MAP.
        for name, attribute_type in self.ATTRIBUTE_TYPE_MAP.items():
            if attribute_type is AttributeType.BOOL:
                value = bool(int(element.get(name, '0')))
            elif attribute_type is AttributeType.LITERAL:
                for subelement in element:
                    if subelement.tag == 'Literal' and subelement.get('type') == name:
                        value = subelement.text.strip()
                        break
                else:
                    value = ''
            elif attribute_type is AttributeType.STRING:
                value = element.get(name, '')
            elif attribute_type is AttributeType.STRING_LIST:
                value = element.get(name, '').split()

            setattr(self.model, name, value)

    def save(self, output):
        """ Save the model to an output file. """

        # This method must be reimplemented by those adapters that write their
        # own XML element.  However we don't want to make it abstract and have
        # to provide a stub reimplementation in other adapters.
        raise NotImplementedError

    @classmethod
    def save_attribute(cls, name, value, output):
        """ Save an attribute. """

        output.write(f' {name}="{cls._escape(value)}"')

    def save_attributes(self, output):
        """ Save the XML attributes of an adapter that does not write its own
        XML element.
        """

        # This default implementation assumes there are no attributes.
        pass

    def save_bool(self, name, output):
        """ Save a bool. """

        value = getattr(self.model, name)

        if value:
            self.save_attribute(name, '1', output)

    def save_literal(self, name, output):
        """ Save the value of a literal text attribute. """

        value = getattr(self.model, name)

        if value != '':
            output.write(f'<Literal type="{name}">\n{self._escape(value)}\n</Literal>\n', indent=False)

    def save_str(self, name, output):
        """ Save a string. """

        value = getattr(self.model, name)

        if value != '':
            self.save_attribute(name, value, output)

    def save_str_list(self, name, output):
        """ Save a list of strings. """

        value = getattr(self.model, name)

        if len(value) != 0:
            self.save_attribute(name, ' '.join(value), output)

    def save_subelements(self, output):
        """ Save the XML subelements of an adapter that does not write its own
        XML element.
        """

        # This default implementation assumes there are no subelements.
        pass

    @staticmethod
    def _escape(s):
        """ Return an XML escaped string. """

        return escape(s, {'"': '&quot;'})

    @classmethod
    def _normalise_templates(cls, type):
        """ Return the normalised form of any templates. """

        t_start = type.find('<')
        t_end = type.rfind('>')

        if t_start > 0 and t_end > t_start:
            xt = []

            # Note that this doesn't handle nested template arguments properly.
            for t_arg in type[t_start + 1:t_end].split(','):
                xt.append(cls._normalise_templates(t_arg.strip()))

            type = type[:t_start + 1] + ', '.join(xt) + type[t_end:]

        return type


class BaseApiAdapter(BaseAdapter):
    """ This is the base class for all adapters for models that are written to
    a .sip file and provide a user-friendly, one line string representation.
    """

    @abstractmethod
    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        ...

    def version_start(self, output):
        """ Write the start of the version tests for an API.  Returns the
        number of %End statements needed to be passed to the corresponding call
        to version_end().
        """

        api = self.model

        nr_ends = 0

        for vrange in api.versions:
            vr = version_range(vrange)
            output.write(f'%If ({vr})\n', indent=False)
            nr_ends += 1

        # Multiple platforms are logically or-ed.
        if len(api.platforms) != 0:
            platforms = ' || '.join(api.platforms)
            output.write(f'%If ({platforms})\n', indent=False)
            nr_ends += 1

        # Multiple features are nested (ie. logically and-ed).
        for feature in api.features:
            output.write(f'%If ({feature})\n', indent=False)
            nr_ends += 1

        # Also handle comments.
        if api.comments != '':
            for line in api.comments.split('\n'):
                line = line.rstrip()
                if line != '':
                    line = '// ' + line + '\n'
                else:
                    line = '//\n'

                output.write(line)

        return nr_ends

    @staticmethod
    def version_end(nr_ends, output):
        """ Write the end of the version tests for an API item. """

        for _ in range(nr_ends):
            output.write('%End\n', indent=False)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..access import Access
from ..annos import Annos
from ..argument import Argument
from ..callable import Callable
from ..code import Code
from ..code_container import CodeContainer
from ..constructor import Constructor
from ..destructor import Destructor
from ..docstring import Docstring
from ..enum import Enum
from ..enum_value import EnumValue
from ..extended_access import ExtendedAccess
from ..function import Function
from ..header_directory import HeaderDirectory
from ..header_file import HeaderFile
from ..header_file_version import HeaderFileVersion
from ..klass import Class
from ..manual_code import ManualCode
from ..method import Method
from ..module import Module
from ..namespace import Namespace
from ..opaque_class import OpaqueClass
from ..operator_cast import OperatorCast
from ..operator_function import OperatorFunction
from ..operator_method import OperatorMethod
from ..platform import Platform
from ..project import Project
from ..sip_file import SipFile
from ..tagged import Tagged
from ..typedef import Typedef
from ..variable import Variable
from ..workflow import Workflow

from .access import AccessAdapter
from .annos import AnnosAdapter
from .argument import ArgumentAdapter
from .callable import CallableAdapter
from .code import CodeAdapter
from .code_container import CodeContainerAdapter
from .constructor import ConstructorAdapter
from .destructor import DestructorAdapter
from .docstring import DocstringAdapter
from .enum import EnumAdapter
from .enum_value import EnumValueAdapter
from .extended_access import ExtendedAccessAdapter
from .function import FunctionAdapter
from .header_directory import HeaderDirectoryAdapter
from .header_file import HeaderFileAdapter
from .header_file_version import HeaderFileVersionAdapter
from .klass import ClassAdapter
from .manual_code import ManualCodeAdapter
from .method import MethodAdapter
from .module import ModuleAdapter
from .namespace import NamespaceAdapter
from .opaque_class import OpaqueClassAdapter
from .operator_cast import OperatorCastAdapter
from .operator_function import OperatorFunctionAdapter
from .operator_method import OperatorMethodAdapter
from .platform import PlatformAdapter
from .project import ProjectAdapter
from .sip_file import SipFileAdapter
from .tagged import TaggedAdapter
from .typedef import TypedefAdapter
from .variable import VariableAdapter
from .workflow import WorkflowAdapter


# The map of adaptable models to adapters.
ADAPTER_MAP = {
    Access:             AccessAdapter,
    Annos:              AnnosAdapter,
    Argument:           ArgumentAdapter,
    Callable:           CallableAdapter,
    Class:              ClassAdapter,
    Code:               CodeAdapter,
    CodeContainer:      CodeContainerAdapter,
    Constructor:        ConstructorAdapter,
    Destructor:         DestructorAdapter,
    Docstring:          DocstringAdapter,
    Enum:               EnumAdapter,
    EnumValue:          EnumValueAdapter,
    ExtendedAccess:     ExtendedAccessAdapter,
    Function:           FunctionAdapter,
    HeaderDirectory:    HeaderDirectoryAdapter,
    HeaderFile:         HeaderFileAdapter,
    HeaderFileVersion:  HeaderFileVersionAdapter,
    ManualCode:         ManualCodeAdapter,
    Method:             MethodAdapter,
    Module:             ModuleAdapter,
    Namespace:          NamespaceAdapter,
    OperatorCast:       OperatorCastAdapter,
    OperatorFunction:   OperatorFunctionAdapter,
    OperatorMethod:     OperatorMethodAdapter,
    OpaqueClass:        OpaqueClassAdapter,
    Platform:           PlatformAdapter,
    Project:            ProjectAdapter,
    SipFile:            SipFileAdapter,
    Tagged:             TaggedAdapter,
    Typedef:            TypedefAdapter,
    Variable:           VariableAdapter,
    Workflow:           WorkflowAdapter,
}



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import re

from ...models import CodeContainer, Enum

from ..helpers import warning


# The regular expression used to validate an identifier.
_identifier_re = re.compile(r'[_A-Z][_A-Z0-9]*', re.ASCII|re.IGNORECASE)

def validate_identifier(identifier, identifier_type):
    """ Validate an identifier and return an error message describing why it is
    invalid, or None if it is valid.
    """

    if identifier == '':
        return f"A {identifier_type} name is required."

    if not _identifier_re.match(identifier):
        return f"A {identifier_type} name can only contain underscores, ASCII letters and digits and cannot start with a digit."

    return None


def validation_error(message, dialog):
    """ Display a validation error message from a dialog. """

    warning(dialog.widget.windowTitle(), message, parent=dialog.widget)


def tagged_items(project):
    """ A generator of 2-tuples of all API items that can have a tag (feature,
    platform or version) and the API item that contains it.  The values are in
    depth first order.
    """

    for module in project.modules:
        for sip_file in module.content:
            for item in _tagged_from_container(sip_file):
                yield item


def _tagged_from_container(container):
    """ A sub-generator for the items in a container. """

    for code in container.content:
        # Depth first.
        if isinstance(code, Enum):
            for enum_value in code.content:
                yield (enum_value, code)
        elif isinstance(code, CodeContainer):
            for item in _tagged_from_container(code):
                yield item

        yield (code, container)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QCheckBox, QFormLayout, QLineEdit

from ...helpers import BaseDialog
from ...shell import EventType

from .helpers import validate_feature_name


class NewFeatureDialog(BaseDialog):
    """ This class implements the dialog for creating a new feature. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        form = QFormLayout()
        layout.addLayout(form)

        self._feature = QLineEdit()
        form.addRow("Feature name", self._feature)

        self._external = QCheckBox("Feature is defined in another project?")
        layout.addWidget(self._external)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        feature = self._feature.text().strip()
        if not validate_feature_name(feature, project, self):
            return False

        if self._external.isChecked():
            features_list = project.externalfeatures
        else:
            features_list = project.features

        features_list.append(feature)

        self.shell.notify(EventType.FEATURE_ADD, feature)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass, field


@dataclass
class CodeContainer:
    """ This class is a mixin for APIs that can contain other APIs. """

    # The list of contained API items.
    content: list['.code.Code'] = field(default_factory=list)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .access import Access
from .callable import Callable
from .docstring import Docstring


@dataclass
class Constructor(Callable, Docstring, Access):
    """ This class implements a C++ constructor. """

    # This is set if the constructor is explicit.
    explicit: bool = False



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .access import Access
from .code import Code


@dataclass
class Variable(Code, Access):
    """ This class implements C struct and C++ class member variables. """

    # The optional %AccessCode.
    accesscode: str = ''

    # The optional %GetCode.
    getcode: str = ''

    # The name of the variable.
    name: str = ''

    # The optional %SetCode.
    setcode: str = ''

    # This is set if the variable is static.
    static: bool = False

    # The type of the variable.
    type: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtGui import QAction

from ...shell import EventType
from ...shell_tool import ActionTool


class PlatformsTool(ActionTool):
    """ This class implements the platforms tool. """

    @property
    def actions(self):
        """ Get the destination menu and sequence of actions handled by the
        tool.
        """

        self._new_action = QAction("New Platform...",
                triggered=self._handle_new)
        self._rename_action = QAction("Rename Platform...",
                triggered=self._handle_rename)
        self._delete_action = QAction("Delete Platform...",
                triggered=self._handle_delete)

        return ("Edit",
                (self._new_action, self._rename_action, self._delete_action))

    def event(self, event_type, event_arg):
        """ Reimplemented to handle project-specific events. """

        if event_type in (EventType.PROJECT_NEW, EventType.PLATFORM_ADD, EventType.PLATFORM_DELETE):
            # Configure the actions.
            are_platforms = (len(self.shell.project.platforms) != 0)
            self._rename_action.setEnabled(are_platforms)
            self._delete_action.setEnabled(are_platforms)

    @property
    def name(self):
        """ Get the tool's name. """

        return 'metasip.tool.platforms'

    def _handle_delete(self):
        """ Handle the Delete action. """

        from .delete_platform_dialog import DeletePlatformDialog

        self.shell.handle_project_dialog("Delete Platform",
                DeletePlatformDialog)

    def _handle_new(self):
        """ Handle the New action. """

        from .new_platform_dialog import NewPlatformDialog

        self.shell.handle_project_dialog("New Platform", NewPlatformDialog)

    def _handle_rename(self):
        """ Handle the Rename action. """

        from .rename_platform_dialog import RenamePlatformDialog

        self.shell.handle_project_dialog("Rename Platform",
                RenamePlatformDialog)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .annos import Annos


@dataclass
class Argument(Annos):
    """ This class implements an argument of a callable. """

    # The optional C/C++ default value of the argument.  This is interpreted
    # literally so that (for example) a string value includes the quotes.
    default: str = ''

    # The optional name of the argument.
    name: str = ''

    # The optional Python default value of the argument.
    pydefault: str = ''

    # The optional Python type of the argument.
    pytype: str = ''

    # The C/C++ type of the argument.
    type: str = ''

    # This is set if the name of the argument hasn't been verified, i.e.
    # accepted for use (typically as a keyword argument).
    unnamed: bool = True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .abstract_project_ui import AbstractProjectUi
from .generate_sip_files import generate_sip_files
from .load_project import load_project
from .save_project import save_project



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..access import Access
from ..annos import Annos
from ..constructor import Constructor
from ..destructor import Destructor
from ..docstring import Docstring
from ..extended_access import ExtendedAccess
from ..klass import Class
from ..code import Code
from ..code_container import CodeContainer
from ..enum import Enum
from ..manual_code import ManualCode
from ..method import Method
from ..namespace import Namespace
from ..opaque_class import OpaqueClass
from ..operator_cast import OperatorCast
from ..operator_method import OperatorMethod
from ..typedef import Typedef
from ..variable import Variable

from .adapt import adapt
from .base_adapter import AttributeType, BaseApiAdapter


class ClassAdapter(BaseApiAdapter):
    """ This is the Class adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'bases':            AttributeType.STRING,
        'bicharbufcode':    AttributeType.LITERAL,
        'bigetbufcode':     AttributeType.LITERAL,
        'bireadbufcode':    AttributeType.LITERAL,
        'birelbufcode':     AttributeType.LITERAL,
        'bisegcountcode':   AttributeType.LITERAL,
        'biwritebufcode':   AttributeType.LITERAL,
        'convfromtypecode': AttributeType.LITERAL,
        'convtotypecode':   AttributeType.LITERAL,
        'gcclearcode':      AttributeType.LITERAL,
        'gctraversecode':   AttributeType.LITERAL,
        'name':             AttributeType.STRING,
        'picklecode':       AttributeType.LITERAL,
        'pybases':          AttributeType.STRING,
        'struct':           AttributeType.BOOL,
        'subclasscode':     AttributeType.LITERAL,
        'finalisationcode': AttributeType.LITERAL,
        'typecode':         AttributeType.LITERAL,
        'typeheadercode':   AttributeType.LITERAL,
        'typehintcode':     AttributeType.LITERAL,
    }

    # The map of element tags and Code sub-class factories.
    _TAG_CODE_MAP = {
        'Class':            Class,
        'Constructor':      Constructor,
        'Destructor':       Destructor,
        'Enum':             Enum,
        'ManualCode':       ManualCode,
        'Method':           Method,
        'Namespace':        Namespace,
        'OpaqueClass':      OpaqueClass,
        'OperatorCast':     OperatorCast,
        'OperatorMethod':   OperatorMethod,
        'Typedef':          Typedef,
        'Variable':         Variable,
    }

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        klass = self.model
        other_klass = other.model

        if type(klass) is not type(other_klass):
            return False

        if klass.access != other_klass.access:
            return False

        if klass.name != other_klass.name:
            return False

        if klass.struct != other_klass.struct:
            return False

        if klass.bases != other_klass.bases:
            return False

        return True

    def as_str(self):
        """ Return the standard string representation. """

        klass = self.model

        s = 'struct' if klass.struct else 'class'

        if klass.name != '':
            s += ' ' + klass.name

        if klass.bases != '':
            s += ' : ' + klass.bases

        s += adapt(klass, Annos).as_str()

        return s

    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        klass = self.model

        nr_ends = self.version_start(output)

        output.blank()

        type_type = 'struct ' if klass.struct else 'class '
        bases_s = ''

        if klass.pybases != '':
            # Treat None as meaning no super classes:
            if klass.pybases != 'None':
                bases_s = ' : ' + ', '.join(klass.pybases.split())
        elif klass.bases != '':
            bases = []

            for base in klass.bases.split(', '):
                access, base_cls = base.split()
                bases.append(f'{access} {base_cls}')

            bases_s = ' : ' + ', '.join(bases)

        output.write(type_type + klass.name + bases_s + adapt(klass, Annos).as_str() + '\n{\n')

        adapt(klass, Docstring).generate_sip_directives(output)
        output.write_code_directive('%TypeHintCode', klass.typehintcode,
                indent=False)

        output.write('%TypeHeaderCode\n', indent=False)

        if klass.typeheadercode != '':
            output.write(klass.typeheadercode + '\n', indent=False)
        else:
            output.write(f'#include <{sip_file.name}>\n', indent=False)

        output.write('%End\n', indent=False)

        output.blank()

        output.write_code_directive('%TypeCode', klass.typecode, indent=False)
        output.write_code_directive('%FinalisationCode',
                klass.finalisationcode)
        output.write_code_directive('%ConvertToSubClassCode',
                klass.subclasscode)
        output.write_code_directive('%ConvertToTypeCode', klass.convtotypecode,
                indent=False)
        output.write_code_directive('%ConvertFromTypeCode',
                klass.convfromtypecode, indent=False)
        output.write_code_directive('%GCTraverseCode', klass.gctraversecode)
        output.write_code_directive('%GCClearCode', klass.gcclearcode)
        output.write_code_directive('%BIGetBufferCode', klass.bigetbufcode)
        output.write_code_directive('%BIReleaseBufferCode', klass.birelbufcode)
        output.write_code_directive('%BIGetReadBufferCode',
                klass.bireadbufcode)
        output.write_code_directive('%BIGetWriteBufferCode',
                klass.biwritebufcode)
        output.write_code_directive('%BIGetSegCountCode', klass.bisegcountcode)
        output.write_code_directive('%BIGetCharBufferCode',
                klass.bicharbufcode)
        output.write_code_directive('%PickleCode', klass.picklecode)

        output += 1

        access = '' if klass.struct else 'private'

        for api in klass.content:
            if api.status != '':
                continue

            if isinstance(api, (Access, ExtendedAccess)):
                if access != api.access:
                    output -= 1
                    access = api.access

                    if access != '':
                        access_s = access
                    else:
                        access_s = 'public'

                    output.blank()
                    output.write(access_s + ':\n')
                    output += 1

            adapt(api).generate_sip(sip_file, output)

        output -= 1

        output.write('};\n')

        output.blank()

        self.version_end(nr_ends, output)

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        adapt(self.model, Code).load(element, project, ui)
        adapt(self.model, CodeContainer).load(self._TAG_CODE_MAP, element,
                project, ui)
        adapt(self.model, Docstring).load(element, project, ui)
        adapt(self.model, Access).load(element, project, ui)

    def save(self, output):
        """ Save the model to an output file. """

        klass = self.model

        output.write('<Class')
        adapt(klass, Code).save_attributes(output)
        adapt(klass, CodeContainer).save_attributes(output)
        adapt(klass, Docstring).save_attributes(output)
        adapt(klass, Access).save_attributes(output)
        self.save_attribute('name', klass.name, output)
        self.save_str('bases', output)
        self.save_str('pybases', output)
        self.save_bool('struct', output)
        output.write('>\n')

        output += 1
        # The order is to match older versions.
        adapt(klass, Code).save_subelements(output)
        adapt(klass, Docstring).save_subelements(output)
        self.save_literal('typehintcode', output)
        self.save_literal('typeheadercode', output)
        self.save_literal('typecode', output)
        self.save_literal('finalisationcode', output)
        self.save_literal('subclasscode', output)
        self.save_literal('convtotypecode', output)
        self.save_literal('convfromtypecode', output)
        self.save_literal('gctraversecode', output)
        self.save_literal('gcclearcode', output)
        self.save_literal('bigetbufcode', output)
        self.save_literal('birelbufcode', output)
        self.save_literal('bireadbufcode', output)
        self.save_literal('biwritebufcode', output)
        self.save_literal('bisegcountcode', output)
        self.save_literal('bicharbufcode', output)
        self.save_literal('picklecode', output)
        adapt(klass, CodeContainer).save_subelements(output)
        adapt(klass, Access).save_subelements(output)
        output -= 1

        output.write('</Class>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..helpers import validate_identifier, validation_error


def init_feature_selector(selector, project):
    """ Initialise a feature selector. """

    selector.clear()
    selector.addItems(sorted(project.features + project.externalfeatures))


def validate_feature_name(feature, project, dialog):
    """ Validate a feature name and return True if it is valid. """

    message = validate_identifier(feature, "feature")
    if message is None:
        if feature in project.features:
            message = "A feature has already been defined with the same name."
        elif feature in project.externalfeatures:
            message = "An external feature has already been defined with the same name."

    if message is not None:
        validation_error(message, dialog)
        return False

    return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QComboBox, QFormLayout, QLineEdit

from ...helpers import BaseDialog
from ...shell import EventType

from ..helpers import tagged_items

from .helpers import init_platform_selector, validate_platform_name


class RenamePlatformDialog(BaseDialog):
    """ This class implements the dialog for renaming a platform. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        self._platform = QComboBox()
        layout.addWidget(self._platform)

        form = QFormLayout()
        layout.addLayout(form)

        self._new_name = QLineEdit()
        form.addRow("New name", self._new_name)

    def set_fields(self):
        """ Set the dialog's fields from the project. """

        init_platform_selector(self._platform, self.model)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        old_name = self._platform.currentText()

        new_name = self._new_name.text().strip()
        if not validate_platform_name(new_name, project, self):
            return False

        # Rename in each API item it appears.
        for api_item, _ in tagged_items(project):
            for i, p in enumerate(api_item.platforms):
                if p[0] == '!':
                    if p[1:] == old_name:
                        api_item.platforms[i] = '!' + new_name
                elif p == old_name:
                    api_item.platforms[i] = new_name

        # Rename in the project's list.
        project.platforms[project.platforms.index(old_name)] = new_name

        self.shell.notify(EventType.PLATFORM_RENAME, (old_name, new_name))

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtGui import QAction

from ...helpers import question
from ...shell import EventType
from ...shell_tool import ActionTool

from .helpers import delete_version


class VersionsTool(ActionTool):
    """ This class implements the versions tool. """

    @property
    def actions(self):
        """ Get the destination menu and sequence of actions handled by the
        tool.
        """

        self._new_action = QAction("New Version...",
                triggered=self._handle_new)
        self._rename_action = QAction("Rename Version...",
                triggered=self._handle_rename)
        self._delete_action = QAction("Delete Version...",
                triggered=self._handle_delete)
        self._delete_all_action = QAction("Delete All Versions...",
                triggered=self._handle_delete_all)

        return ("Edit",
                (self._new_action, self._rename_action, self._delete_action,
                        self._delete_all_action))

    def event(self, event_type, event_arg):
        """ Reimplemented to handle project-specific events. """

        if event_type in (EventType.PROJECT_NEW, EventType.VERSION_ADD, EventType.VERSION_DELETE):
            # Configure the actions.
            are_versions = (len(self.shell.project.versions) != 0)
            self._rename_action.setEnabled(are_versions)
            self._delete_action.setEnabled(are_versions)
            self._delete_all_action.setEnabled(are_versions)

    @property
    def name(self):
        """ Get the tool's name. """

        return 'metasip.tool.versions'

    def _handle_delete(self):
        """ Handle the Delete action. """

        from .delete_version_dialog import DeleteVersionDialog

        self.shell.handle_project_dialog("Delete Version", DeleteVersionDialog)

    def _handle_delete_all(self):
        """ Handle the Delete All action. """

        confirmed = question("Delete All Versions",
                "All versions will be removed along with any API items that "
                "are not part of the latest version.\n\nDo you wish to "
                "continue?",
                parent=self.shell.widget)

        if confirmed:
            for version in list(self.shell.project.versions):
                delete_version(version, self.shell, migrate_items=False)

            self.shell.dirty = True

    def _handle_new(self):
        """ Handle the New action. """

        from .new_version_dialog import NewVersionDialog

        self.shell.handle_project_dialog("New Version", NewVersionDialog)

    def _handle_rename(self):
        """ Handle the Rename action. """

        from .rename_version_dialog import RenameVersionDialog

        self.shell.handle_project_dialog("Rename Version", RenameVersionDialog)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .annos import Annos
from .tagged import Tagged
from .workflow import Workflow


@dataclass
class EnumValue(Annos, Tagged, Workflow):
    """ This class implements an enum value. """

    # The name of the value.
    name: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QMessageBox


def warning(title, text, detail=None, parent=None):
    """ Display a warning message to the user. """

    message_box = QMessageBox(QMessageBox.Icon.Warning, title, text,
            parent=parent)

    if detail:
        message_box.setDetailedText(detail)

    message_box.exec()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import (QCheckBox, QComboBox, QFormLayout, QGroupBox,
        QLineEdit, QVBoxLayout)

from ....helpers import BaseDialog

from .helpers import Encoding, split_annos


class VariablePropertiesDialog(BaseDialog):
    """ This class implements the dialog for a variables's properties. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        group_box = QGroupBox("Annotations")
        layout.addWidget(group_box)
        group_box_layout = QVBoxLayout()
        group_box.setLayout(group_box_layout)

        form = QFormLayout()
        group_box_layout.addLayout(form)

        self._doc_type = QLineEdit()
        form.addRow('DocType', self._doc_type)

        encoding = QComboBox()
        self._encoding_helper = Encoding(encoding)
        form.addRow('Encoding', encoding)

        self._no_setter = QCheckBox('NoSetter')
        group_box_layout.addWidget(self._no_setter)

        self._no_type_hint = QCheckBox('NoTypeHint')
        group_box_layout.addWidget(self._no_type_hint)

        self._py_int = QCheckBox('PyInt')
        group_box_layout.addWidget(self._py_int)

        form = QFormLayout()
        group_box_layout.addLayout(form)

        self._py_name = QLineEdit()
        form.addRow('PyName', self._py_name)

        self._type_hint = QLineEdit()
        form.addRow('TypeHint', self._type_hint)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        for name, value in split_annos(self.model.annos):
            if name == 'DocType':
                self._doc_type.setText(value)
            elif name == 'Encoding':
                self._encoding_helper.setAnnotation(value)
            elif name == 'NoSetter':
                self._no_setter.setChecked(True)
            elif name == 'NoTypeHint':
                self._no_type_hint.setChecked(True)
            elif name == 'PyInt':
                self._py_int.setChecked(True)
            elif name == 'PyName':
                self._py_name.setText(value)
            elif name == 'TypeHint':
                self._type_hint.setText(value)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        annos_list = []

        self._encoding_helper.annotation(annos_list)

        doc_type = self._doc_type.text().strip()
        if doc_type:
            annos_list.append(f'DocType="{doc_type}"')

        if self._no_setter.isChecked():
            annos_list.append('NoSetter')

        if self._no_type_hint.isChecked():
            annos_list.append('NoTypeHint')

        if self._py_int.isChecked():
            annos_list.append('PyInt')

        py_name = self._py_name.text().strip()
        if py_name:
            annos_list.append('PyName=' + py_name)

        type_hint = self._type_hint.text().strip()
        if type_hint:
            annos_list.append(f'TypeHint="{type_hint}"')

        self.model.annos = ','.join(annos_list)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import (QCheckBox, QComboBox, QFormLayout, QGridLayout,
        QGroupBox, QHBoxLayout, QLabel, QLineEdit)

from ....helpers import BaseDialog

from .helpers import Encoding, split_annos


class ArgumentPropertiesDialog(BaseDialog):
    """ This class implements the dialog for argument properties. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        name_layout = QHBoxLayout()
        layout.addLayout(name_layout)

        name_layout.addWidget(QLabel("Name"))
        self._name = QLineEdit()
        name_layout.addWidget(self._name)
        self._unnamed = QCheckBox("Unnamed")
        name_layout.addWidget(self._unnamed)

        form = QFormLayout()
        layout.addLayout(form)

        self._py_type = QLineEdit()
        form.addRow("Python type", self._py_type)

        self._py_default = QLineEdit()
        form.addRow("Python default value", self._py_default)

        group_box = QGroupBox("Annotations")
        layout.addWidget(group_box)
        group_box_layout = QGridLayout()
        group_box.setLayout(group_box_layout)

        self._allow_none = QCheckBox('AllowNone')
        group_box_layout.addWidget(self._allow_none, 0, 0)

        self._array = QCheckBox('Array')
        group_box_layout.addWidget(self._array, 1, 0)

        self._array_size = QCheckBox('ArraySize')
        group_box_layout.addWidget(self._array_size, 2, 0)

        self._constrained = QCheckBox('Constrained')
        group_box_layout.addWidget(self._constrained, 3, 0)

        self._disallow_none = QCheckBox('DisallowNone')
        group_box_layout.addWidget(self._disallow_none, 4, 0)

        form = QFormLayout()
        group_box_layout.addLayout(form, 5, 0, 4, 1)

        self._doc_type = QLineEdit()
        form.addRow('DocType', self._doc_type)

        self._doc_value = QLineEdit()
        form.addRow('DocValue', self._doc_value)

        self._doc_type = QLineEdit()
        form.addRow('DocType', self._doc_type)

        encoding = QComboBox()
        form.addRow('Encoding', encoding)
        self._encoding_helper = Encoding(encoding)

        self._get_wrapper = QCheckBox('GetWrapper')
        group_box_layout.addWidget(self._get_wrapper, 9, 0)

        self._in = QCheckBox('In')
        group_box_layout.addWidget(self._in, 10, 0)

        self._keep_reference = QCheckBox('KeepReference')
        group_box_layout.addWidget(self._keep_reference, 11, 0)
        self._keep_reference.stateChanged.connect(self._update_reference)

        self._no_copy = QCheckBox('NoCopy')
        group_box_layout.addWidget(self._no_copy, 12, 0)

        self._out = QCheckBox('Out')
        group_box_layout.addWidget(self._out, 0, 1)

        self._py_int = QCheckBox('PyInt')
        group_box_layout.addWidget(self._py_int, 1, 1)

        form = QFormLayout()
        group_box_layout.addLayout(form, 2, 1)

        self._reference_label = QLabel('Reference')
        self._reference = QLineEdit()
        form.addRow(self._reference_label, self._reference)

        self._result_size = QCheckBox('ResultSize')
        group_box_layout.addWidget(self._result_size, 3, 1)

        form = QFormLayout()
        group_box_layout.addLayout(form, 4, 1)

        self._scopes_stripped = QLineEdit()
        form.addRow('ScopesStripped', self._scopes_stripped)

        self._single_shot = QCheckBox('SingleShot')
        group_box_layout.addWidget(self._single_shot, 5, 1)

        self._transfer = QCheckBox('Transfer')
        group_box_layout.addWidget(self._transfer, 6, 1)

        self._transfer_back = QCheckBox('TransferBack')
        group_box_layout.addWidget(self._transfer_back, 7, 1)

        self._transfer_this = QCheckBox('TransferThis')
        group_box_layout.addWidget(self._transfer_this, 8, 1)

        form = QFormLayout()
        group_box_layout.addLayout(form, 9, 1, 4, 1)

        self._type_hint = QLineEdit()
        form.addRow('TypeHint', self._type_hint)

        self._type_hint_in = QLineEdit()
        form.addRow('TypeHintIn', self._type_hint_in)

        self._type_hint_out = QLineEdit()
        form.addRow('TypeHintOut', self._type_hint_out)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        self._name.setText(self.model.name)
        self._unnamed.setChecked(self.model.unnamed)
        self._py_type.setText(self.model.pytype)
        self._py_default.setText(self.model.pydefault)

        for name, value in split_annos(self.model.annos):
            if name == 'AllowNone':
                self._allow_none.setChecked(True)
            elif name == 'Array':
                self._array.setChecked(True)
            elif name == 'ArraySize':
                self._array_size.setChecked(True)
            elif name == 'Constrained':
                self._constrained.setChecked(True)
            elif name == 'DisallowNone':
                self._disallow_none.setChecked(True)
            elif name == 'DocType':
                self._doc_type.setText(value)
            elif name == 'DocValue':
                self._doc_value.setText(value)
            elif name == 'Encoding':
                self._encoding_helper.setAnnotation(value)
            elif name == 'GetWrapper':
                self._get_wrapper.setChecked(True)
            elif name == 'In':
                self._in.setChecked(True)
            elif name == 'KeepReference':
                self._keep_reference.setChecked(True)

                if value is not None:
                    self._reference.setText(value)
            elif name == 'NoCopy':
                self._no_copy.setChecked(True)
            elif name == 'Out':
                self._out.setChecked(True)
            elif name == 'PyInt':
                self._py_int.setChecked(True)
            elif name == 'ResultSize':
                self._result_size.setChecked(True)
            elif name == 'SingleShot':
                self._single_shot.setChecked(True)
            elif name == 'ScopesStripped':
                self._scopes_stripped.setText(value)
            elif name == 'Transfer':
                self._transfer.setChecked(True)
            elif name == 'TransferBack':
                self._transfer_back.setChecked(True)
            elif name == 'TransferThis':
                self._transfer_this.setChecked(True)
            elif name == 'TypeHint':
                self._type_hint.setText(value)
            elif name == 'TypeHintIn':
                self._type_hint_in.setText(value)
            elif name == 'TypeHintOut':
                self._type_hint_out.setText(value)

        self._update_reference()

    def get_fields(self):
        """ Update the API item from the dialog's field. """

        self.model.name = self._name.text().strip()
        self.model.unnamed = self._unnamed.isChecked()
        self.model.pytype = self._py_type.text().strip()
        self.model.pydefault = self._py_default.text().strip()

        annos_list = []

        if self._allow_none.isChecked():
            annos_list.append('AllowNone')

        if self._array.isChecked():
            annos_list.append('Array')

        if self._array_size.isChecked():
            annos_list.append('ArraySize')

        if self._constrained.isChecked():
            annos_list.append('Constrained')

        if self._disallow_none.isChecked():
            annos_list.append('DisallowNone')

        doc_type = self._doc_type.text().strip()
        if doc_type:
            annos_list.append(f'DocType="{doc_type}"')

        doc_value = self._doc_value.text().strip()
        if doc_value:
            annos_list.append(f'DocValue="{doc_value}"')

        self._encoding_helper.annotation(annos_list)

        if self._get_wrapper.isChecked():
            annos_list.append('GetWrapper')

        if self._in.isChecked():
            annos_list.append('In')

        if self._keep_reference.isChecked():
            reference = self.ref.text().strip()
            if reference:
                annos_list.append('KeepReference=' + reference)
            else:
                annos_list.append('KeepReference')

        if self._no_copy.isChecked():
            annos_list.append('NoCopy')

        if self._out.isChecked():
            annos_list.append('Out')

        if self._py_int.isChecked():
            annos_list.append('PyInt')

        if self._result_size.isChecked():
            annos_list.append('ResultSize')

        if self._single_shot.isChecked():
            annos_list.append('SingleShot')

        scopes_stripped = self._scopes_stripped.text().strip()
        if scopes_stripped:
            annos_list.append('ScopesStripped=' + scopes_stripped)

        if self._transfer.isChecked():
            annos_list.append('Transfer')

        if self._transfer_back.isChecked():
            annos_list.append('TransferBack')

        if self._transfer_this.isChecked():
            annos_list.append('TransferThis')

        type_hint = self._type_hint.text().strip()
        if type_hint:
            annos_list.append(f'TypeHint="{type_hint}"')

        type_hint_in = self._type_hint_in.text().strip()
        if type_hint_in:
            annos_list.append(f'TypeHintIn="{type_hint_in}"')

        type_hint_out = self._type_hint_out.text().strip()
        if type_hint_out:
            annos_list.append(f'TypeHintOut="{type_hint_out}"')

        self.model.annos = ','.join(annos_list)

        return True

    def _update_reference(self, state=None):
        """ Enable the reference field if the keep reference field is checked.
        """

        checked = self._keep_reference.isChecked()

        self._reference_label.setEnabled(checked)
        self._reference.setEnabled(checked)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass, field

from .argument import Argument
from .code import Code


@dataclass
class Callable(Code):
    """ This class implements API items that represent most callables (i.e.
    everything except destructors).
    """

    # The C/C++ arguments.
    args: list[Argument] = field(default_factory=list)

    # The optional %MethodCode.
    methcode: str = ''

    # The name of the callable.
    name: str = ''

    # The optional Python arguments.
    pyargs: str = ''

    # The optional Python return type.
    pytype: str = ''

    # The C/C++ return type.  This is not used by constructors.
    rtype: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass


@dataclass
class Access:
    """ This class is a mixin for API models subject to standard C++ access
    specifiers.
    """

    # The access specifier.  Values are '' (meaning public), 'protected' and
    # 'private'.
    access: str = ''
